"""
Defines data structures useful for the pre-whitening process, with useful self-contained behaviour.

The Lightcurve and Periodogram classes are useful for storing representations of the data under analysis. For example,
the Lightcurve class stores a time series and automatically generates a Periodogram object, representing a pseudo-
amplitude spectrum corresponding to the Lightcurve object. The Frequency class provides a full description of a
single-frequency variability model, including the frequency, amplitude, and phase parameters as well as the sinusoidal
model that was used to optimize it.

Typical workflow is to import a timeseries to a Lightcurve object, identify a
candidate frequency/amplitude pair using its Periodogram object, perform and optimization (see pywhiten.optimization),
and store the results in a Frequency object. If performing iterative pre-whitening to identify multiple frequencies,
it may also be useful to store all these values in a FrequencyContainer object. In most cases where the pre-whitening is
conducted manually, a regular list of Frequency objects will suffice.

Classes:
    Frequency
    FrequencyContainer
    Lightcurve
    Periodogram
"""

from pywhiten.data.Frequency import Frequency
from pywhiten.data.FrequencyContainer import FrequencyContainer
from pywhiten.data.Lightcurve import Lightcurve
from pywhiten.data.Periodogram import Periodogram


import numpy as np
from pywhiten.data.InvalidConstructorArgumentsError import InvalidContructorArgumentsError
from pywhiten.data.Periodogram import Periodogram
import matplotlib.pyplot as pl

class Lightcurve():
    """
    An object storing a time series.
    Attributes:
        time : np.array
            Time indexes of measurements
        data : np.array
            Measurements indexed by the time array
        err : np.array
            Measurement uncertainties on data
        periodogram : Periodogram
            The object storing the frequency spectrum of this time series
    """
    time = None
    data = None
    err = None
    periodogram = None


    def __init__(self, time, data, err = None, cfg = None):
        if (type(time) not in [list, np.ndarray, tuple]) or (type(data) not in [list, np.ndarray, tuple]):
            raise InvalidContructorArgumentsError("Lightcurve initialized with time or data not of type list, tuple,"
                                                  "or ndarray")
        else:
            self.time = np.array(time)
            self.data = np.array(data)

        if err is None :
            self.err = np.ones(len(time))
        elif type(err) in [list, np.ndarray, tuple]:
            self.err = np.array(err)
        else:
            raise InvalidContructorArgumentsError("Lightcurve initialized with errors not of type list, tuple, ndarray,"
                                                  "or none")
        if cfg is None:
            self.periodogram = Periodogram(self.time, self.data)
        else:
            self.periodogram = Periodogram(self.time, self.data, cfg=cfg)


    def unpack(self):
        """
        Gets the time, data, and err arrays.
        Returns:
            np.ndarray: the time array
            np.ndarray: the data array
            np.ndarray: the measurement weights
        """
        return self.time, self.data, self.err

    def measure_N_eff(self):
        """
        Get the number of sign changes in the light curve. This is intended for use with differential lightcurves.
        Returns:
            int: the number of sign changes in the light curve.
        """
        sign_change_count = 0
        for i in range(len(self.time)-1):
            if (self.data[i]>0 and self.data[i+1]<0) or (self.data[i]<0 and self.data[i+1] > 0):
                sign_change_count+=1
        return sign_change_count

    def std(self):
        """
        Get standard deviation of light curve data
        Returns:
            float : std of light curve data axis
        """
        return np.std(self.data)

    def t_span(self):
        """
        Get the time span of the lightcurve
        Returns:
            float : max time value minus minimum time value
        """
        return max(self.time) - min(self.time)

    def debug_plot(self):
        pl.plot(self.time, self.data, linestyle=None, marker=".", color="black", markersize = 2)
        pl.show()
        pl.clf()





"""
Handles the relevant optimization for pre-whitening, including fitting sinusoidal single-frequency and multi-frequency
models.

Typical usage is to use the Optimizer object to conduct single-frequency and multi-frequency optimizations at each
iteration of pre-whitening. This is useful as it handles correctly setting up Frequency objects which are returned,
and retains some information between iterations (e.g. the floating zero point parameter). Sinusoidal and periodogram
noise models can also be directly accessed through this module.
Classes:
    Optimizer
Functions:
    chisq
    sin_model
    cos_model
    n_sin_model
    n_cos_model
    n_model_poly
    slf_noise
"""

from pywhiten.optimization.Optimizer import Optimizer
from pywhiten.optimization.models import *



import lmfit as lm
import numpy as np
from pywhiten.optimization.models import sin_model, n_sin_model, cos_model, n_cos_model

def constant_function(x, z):
    return z

class Optimizer:
    """
    Handles chi-squared minimization of single-frequency and multi-frequency sinusoidal models.
    Attributes:
        sf_func (Function): A sinusoidal function taking arguments of the format x:ndarray, f:float, a:float, p:float
            used for single-frequency optimizations and summed for multi-frequency optimizations when using lmfit as the
            optimization engine
        mf_func (Function): A function taking arguments of the format x:ndarray, *pars, where *pars is an arbitrarily
            large set of floats arranged such that it contains a group of frequency guesses followed by a group of
            amplitude guesses followed by a group of phase guesses followed by a guess for the zero point. The
            frequency, amplitude, and phase guesses must be of equal size, meaning *pars should be of length 3n+1 where
            n is some positie integer
        cfg (dict): A configuration dictionary. See pywhitener.cfg for further details.
        c_zp (float): A guess for the zero point used when performing multi-frequency fits. This is stored as an
            attribute so it can be refined over multiple multi-frequency fits as frequencies are added
    """

    sf_func = None
    mf_func = None
    cfg : dict = None
    c_zp : float = 0

    def __init__(self, cfg):
        if type(cfg) == dict:
            self.cfg = cfg

        # set up optimization functions
        if self.cfg["optimization"]["frequency_model_type"] == "sin":
            self.sf_func = sin_model
            self.mf_func = n_sin_model
        elif self.cfg["optimization"]["frequency_model_type"] == "cos":
            self.sf_func = cos_model
            self.mf_func = n_cos_model
        else:
            raise ValueError(f"Configuration optimization.frequency_model_type is set to"
                             f" {self.cfg['optimization']['frequency_model_type']}, must be 'sin' or 'cos'")

        self.c_zp = 0
    def single_frequency_optimization(self, x, data, err, f0, a0, p0, method="lmfit"):
        """
        Optimizes a single sinusoidal model to an x-y dataset
        Args:
            x (ndarray): x values of data to fit to
            data (ndarray): y values of data to fit to
            err (ndarray): data point weights
            f0 (float): guess for frequency
            a0 (float): guess for amplitude
            p0 (float): guess for phase
            method (str): can be one of "lmfit" or "scipy", and defines the package to use to conduct the minimization
                scipy directly implements MINPACK fortran routines, whereas lmfit implements additional functionality
                to handle parameter boundaries with more grace. Only lmfit works right now

        Returns:

        """
        initial_guess = [f0, a0, p0]
        if method == "lmfit":
            # this loop exists so that we can check the results and re-run the fit if certain conditions aren't met
            # the only certain condition at the moment is whether the phase value has moved more than 0.1 from its
            # initial guess
            while True:
                """Optimize a single-frequency sinusoidal model using lmfit"""
                sfmod = lm.Model(self.sf_func)
                sfmod.set_param_hint("f", value=f0, min=f0 * self.cfg["autopw"]["bounds"]["freq_lower_coeff"],
                                     max=f0 * self.cfg["autopw"]["bounds"]["freq_upper_coeff"])
                sfmod.set_param_hint("a", value=a0, min=a0 * self.cfg["autopw"]["bounds"]["amp_lower_coeff"],
                                     max=a0 * self.cfg["autopw"]["bounds"]["amp_upper_coeff"])
                sfmod.set_param_hint("p", value=p0, min=self.cfg["autopw"]["bounds"]["phase_lower"],
                                     max=self.cfg["autopw"]["bounds"]["phase_upper"])


                f_result = sfmod.fit(data=data, weights=err, x=x)

                # if requested, check to see if phase value moved
                if self.cfg["optimization"]["sf_phase_value_check"] and \
                        abs(f_result.best_values["p"] - initial_guess[2]) < 0.1:
                    initial_guess[2] += 0.17
                else:
                    return f_result.best_values["f"], f_result.best_values["a"], \
                        f_result.best_values["p"], f_result.best_fit

    def multi_frequency_optimization(self, x, data, err, freqs):
        """
        Optimizes a multi-frequency model composed of sinusoidal components against an x-y dataset.
        Args:
            x (ndarray): x values of data to fit to
            data (ndarray): y values of data to fit to
            err (ndarray): data weights
            freqs(list): A list of one or more Frequency objects, where the Frequency
                objects f, a, and p attributes are used as initial guesses for the frequencies, amplitudes, and phases,
                and the these values are updated in-place after the optimization is complete (only an evaluated model
                is returned)

        Returns:
            ndarray: the optimized multi-frequency model evaluated at all values of x
        """
        sf_mods = []
        for i in range(len(freqs)):
            freq = freqs[i]
            sfmod = lm.Model(self.sf_func, prefix=f"f{i}")
            sfmod.set_param_hint(f"f{i}f", value=freq.f, min=freq.f * self.cfg["autopw"]["bounds"]["freq_lower_coeff"],
                                 max=freq.f * self.cfg["autopw"]["bounds"]["freq_upper_coeff"])
            sfmod.set_param_hint(f"f{i}a", value=freq.a, min=freq.a * self.cfg["autopw"]["bounds"]["amp_lower_coeff"],
                                 max=freq.a * self.cfg["autopw"]["bounds"]["amp_upper_coeff"])
            sfmod.set_param_hint(f"f{i}p", value=freq.p, min=self.cfg["autopw"]["bounds"]["phase_lower"],
                                 max=self.cfg["autopw"]["bounds"]["phase_upper"])
            sf_mods.append(sfmod)
        zeroptmodel = lm.Model(constant_function, prefix="zp")
        zeroptmodel.set_param_hint(f"zpz", value=self.c_zp)
        if self.cfg["optimization"]["mf_include_zp"]:
            sf_mods.append(zeroptmodel)
        mf_mod = np.sum(sf_mods)
        f_result = mf_mod.fit(data=data, weights=err, x=x)
        # print(f_result.fit_report())

        for i in range(len(freqs)):
            # check boundaries - If any parameters are suspiciously close to a boundary, warn the user
            c_sfmod = sf_mods[i]
            if self.cfg["optimization"]["mf_boundary_warnings"] != 0:
                for ptype in ["f", "a", "p"]:
                    if abs(c_sfmod.param_hints[ptype]["min"] - f_result.best_values[f"f{i}{ptype}"]) \
                            < self.cfg["optimization"]["mf_boundary_warnings"] * f_result.best_values[f"f{i}{ptype}"]:
                        print(f"\t\t WARNING: {ptype} {f_result.best_values[f'f{i}{ptype}']} of f{i} within"
                              f" {self.cfg['optimization']['mf_boundary_warnings'] * 100}% "
                              f"of lower boundary {c_sfmod.param_hints[ptype]['min']}")
                    elif abs(c_sfmod.param_hints[ptype]["max"] - f_result.best_values[f"f{i}{ptype}"]) < \
                            self.cfg['optimization']['mf_boundary_warnings'] * f_result.best_values[f"f{i}{ptype}"]:
                        print(f"\t\t WARNING: {ptype} {f_result.best_values[f'f{i}{ptype}']} of f{i}"
                              f" within {self.cfg['optimization']['mf_boundary_warnings'] * 100}% "
                              f"of upper boundary {c_sfmod.param_hints[ptype]['max']}")
            # update the frequencies in-place
            freqs[i].f = f_result.best_values[f"f{i}f"]
            freqs[i].a = f_result.best_values[f"f{i}a"]
            freqs[i].p = f_result.best_values[f"f{i}p"]
        if self.cfg["optimization"]["mf_include_zp"]:
            if self.cfg["output"]["print_runtime_messages"]:
                print(f"\tMF ZP: {f_result.best_values['zpz']}")
            self.c_zp = f_result.best_values["zpz"]
        return f_result.best_fit



if __name__ == "__main__":
    import os
    import tomli
    from pywhiten.data import Frequency, FrequencyContainer
    default_config = os.getcwd().strip("optimization") + "cfg/default.toml"
    with open(default_config, "rb") as f:
        cfg = tomli.load(f)

    test_Optimizer = Optimizer(cfg=cfg)

    assert test_Optimizer.sf_func == sin_model
    assert test_Optimizer.mf_func == n_sin_model
    assert test_Optimizer.c_zp == 0

    test_x = np.linspace(0, 10, 1000)
    test_y = sin_model(test_x, 1.0, 10, 0.5) + sin_model(test_x, 1.4, 7, 0.4)

    f1 = Frequency(1.1, 9.0, 0.7, t0=0, n=0)
    f2 = Frequency(1.6, 8.1, 0.8, t0=0, n=1)

    test_container = FrequencyContainer(f1, f2)
    model = test_Optimizer.multi_frequency_optimization(test_x, test_y, np.ones(len(test_x)), test_container)


"""
Models used for chi-squared minimization by the Optimizer are defined here.
"""

import numpy as np

def chisq(data, model, err):
    """Measures chi squared for a model array given a data array and weights"""
    return sum(((data-model)/err)**2)

def sin_model(x, f, a, p):
    """Basic sinusoidal model"""
    return a*np.sin(2*np.pi*(f*x+p))

def cos_model(x, f, a, p):
    """Basic (co)sinusoidal model"""
    return a*np.cos(2*np.pi*(f*x+p))

def n_sin_model(x, *params):
    """Sums an arbitrary number of basic sinusoidal models. Number of sinusoids is defined by the number of parameters
    provided, and the *params input should follow the following form: *freqs, *amps, *phases, zp(=zero point)"""
    l = len(params)-1
    nfreqs = len(params)//3
    y = np.zeros(len(x))
    for i in range(l // 3):
        y += sin_model(x, params[i], params[i+nfreqs], params[i+2*nfreqs])
    return y + params[-1]

def n_cos_model(x, *params):
    """Sums an arbitrary number of basic sinusoidal models. Number of sinusoids is defined by the number of parameters
    provided, and the *params input should follow the following form: *freqs, *amps, *phases, zp(=zero point)"""
    l = len(params)-1
    nfreqs = len(params)//3
    y = np.zeros(len(x))
    for i in range(l // 3):
        y += cos_model(x, params[i], params[i+nfreqs], params[i+2*nfreqs])
    return y + params[-1]

def n_model_poly(x, *params):
    """A polynomial model of arbitrary order, determined by the number of input parameters"""
    power = 0
    try:
        out = np.zeros(len(x))
    except TypeError:
        out = 0
    for p in params:
        out += p * (x**power)
        power += 1
    return out

def slf_noise(x, *params):
    """The model presented in Bowman et al. (2019) describing stochastic low-frequency variability in massive stars"""
    # params = [x0, alpha_0, gamma, Cw]
    return params[1] / (1 + (x / params[0]) ** params[2]) + params[3]


import pywhiten
import numpy as np
def test_Lightcurve():
    test_Lightcurve = pywhiten.data.Lightcurve(np.linspace(0, 30, 14000),
                                               np.random.random(14000), err = np.random.random(14000))




import os
import shutil
import numpy as np

import matplotlib.pyplot as pl
from pywhiten.pwio.utilities import flux2mag, flux2mag_e, format_output
from pywhiten.data.FrequencyContainer import FrequencyContainer
from pywhiten.optimization.models import sin_model, cos_model, n_sin_model, n_cos_model


class OutputManager:
    """
        Handles output data/plots and associated formatting for lightcurves, periodograms, and other data from the
        pre-whitening analysis.
        Attributes:
            np.array model_x: An x-array used to make models of variability models, such that they can be plotted over LCs
            
            dict output_dirs: A dictionary of output directories, containing the following keyed values:
                string main_dir: The absolute path to the main output directory
                string pgs_output: The absolute path to the subdirectory where all periododgram data/plots are saved
                string pgs_slf_output: The absolute path to the subdirectory where periodograms are saved with slf fits overlaid
                string pgs_box_avg_output: "" box average profiles overlaid - Currently nothing is saved here
                string pgs_lopoly_output: "" Low-order polynomial fits overlaid, only the final periodogram is currently saved
                string pgs_data_output: The absolute path to the subdirectory where the raw data of each periodogram is saved
                string lcs_output: The absolute path to the subdirectory where all light curve data/plots are saved
                string lcs_sf_output: The absolute path to the subdirectory where light curve plots are saved with single-freq
                    variability models overlaid
                string lcs_mf_output: The absolute path to the subdirectory where light curves are saved with complete
                    variability models overplotted
                string lcs_data_output: The absolute path to the subdirectory where the raw data of residual light curves are
                    saved
                string misc_output: The absolute path where misc data and plots are saved
        """
    cfg: dict

    sf_func = None
    mf_func = None

    output_dirs = {"base": None,
                   "pgs_base": None,
                   "pgs_raw": None,
                   "pgs_slf": None,
                   "pgs_lopoly": None,
                   "pgs_data": None,
                   "lcs_base": None,
                   "lcs_raw": None,
                   "lcs_sf": None,
                   "lcs_mf": None,
                   "lcs_data": None,
                   "auxiliary": None}

    def __init__(self, cfg):
        # Make directories
        try:
            # set up base directory
            if cfg["output"]["paths"]["base"] in [".", "cwd", "/."]:
                self.output_dirs["base"] = os.getcwd()
            else:
                self.output_dirs["base"] = os.getcwd() + cfg["output"]["paths"]["base"]
            # set up second level directories
            for key in ["pgs_base", "lcs_base", "auxiliary"]:
                self.output_dirs[key] = self.output_dirs["base"] + cfg["output"]["paths"][key]
            # set up pgs directories
            for key in ["pgs_raw", "pgs_slf", "pgs_lopoly", "pgs_data"]:
                self.output_dirs[key] = self.output_dirs["pgs_base"] + cfg["output"]["paths"][key]
            # set up lcs directories
            for key in ["lcs_raw", "lcs_sf", "lcs_mf", "lcs_data"]:
                self.output_dirs[key] = self.output_dirs["lcs_base"] + cfg["output"]["paths"][key]
        except KeyError:
            raise KeyError("Issue with specified output directories")

        if os.path.exists(self.output_dirs["base"]):
            shutil.rmtree(self.output_dirs["base"])

        # now actually make the directories
        for path_key in self.output_dirs:
            os.makedirs(self.output_dirs[path_key], exist_ok=True)

        self.cfg = cfg

        # set the model attributes so models can be generated and plotted from the parameters
        if self.cfg["optimization"]["frequency_model_type"] == "sin":
            self.sf_func = sin_model
            self.mf_func = n_sin_model
        elif self.cfg["optimization"]["frequency_model_type"] == "cos":
            self.sf_func = cos_model
            self.mf_func = n_cos_model
        else:
            raise ValueError(f"Configuration optimization.frequency_model_type is set to"
                             f" {self.cfg['optimization']['frequency_model_type']}, must be 'sin' or 'cos'")

    def save_it(self, lcs, frequency_container, zp=0):
        """
        Saves plots and data for a single pre-whitening iteration using the up-to-date light curve list and frequency
        container from the pre-whitening. It assumes the last value in each is from the iteration of interest
        Args:
            lcs (list): a list containing Lightcurve objects
            frequency_container (FrequencyContainer): A frequency container object with non-zero number of frequencies
            zp (float): The zero point

        Returns:

        """
        freqs = frequency_container.get_flist()
        n = len(freqs) - 1
        c_lc = lcs[-1]
        c_pg = c_lc.periodogram
        c_freq = freqs[-1]

        # Save LC, PG data
        self.save_lc(c_lc, f"{self.output_dirs['lcs_data']}/lc{n}.txt")
        self.save_pg(c_pg, f"{self.output_dirs['pgs_data']}/pg{n}.txt")

        # ================= LC PLOTS ===================
        # regular ======================================
        self.add_lc_plot_to_curfig(x=c_lc.time, y=c_lc.data)
        self.add_lc_formatting_to_curfig()
        pl.savefig(f"{self.output_dirs['lcs_raw']}/lc{n}.png")
        pl.clf()
        # sf ===========================================
        self.add_lc_plot_to_curfig(x=c_lc.time, y=c_lc.data, label="Data")
        self.add_lc_plot_to_curfig(x=lcs[0].time, y=c_freq.evaluate_model(lcs[0].time),
                                   color="red", label="SF Model")
        self.add_lc_formatting_to_curfig(legend=True)
        pl.savefig(f"{self.output_dirs['lcs_sf']}/lc_sf{n}.png")
        pl.clf()
        # mf ===========================================
        self.add_lc_plot_to_curfig(x=lcs[0].time, y=lcs[0].data, label="Data")
        mf_y = np.zeros(len(lcs[0].time)) + zp
        for freq in freqs:
            mf_y += freq.evaluate_model(lcs[0].time)
        self.add_lc_plot_to_curfig(x=lcs[0].time, y=mf_y, color="red", label="Complete Variability Model")
        self.add_lc_formatting_to_curfig(legend=True)
        pl.savefig(f"{self.output_dirs['lcs_mf']}/lc_mf{n}.png")
        pl.clf()

        # ================= PG PLOTS ===================
        # regular ======================================
        self.add_pg_plot_to_curfig(x=c_pg.lsfreq, y=c_pg.lsamp, label="Data")
        self.add_pg_formatting_to_curfig()
        pl.savefig(f"{self.output_dirs['pgs_raw']}/pg{n}.png")
        pl.clf()
        # slf ==========================================
        if c_pg.slf_p is not None:
            self.add_pg_plot_to_curfig(x=c_pg.lsfreq, y=c_pg.lsamp, label="Data")
            self.add_pg_plot_to_curfig(x=c_pg.lsfreq, y=c_pg.eval_slf_model(c_pg.lsfreq), color="red", label="SLF Fit")
            self.add_pg_plot_to_curfig(x=c_pg.lsfreq,
                                       y=c_pg.eval_slf_model(c_pg.lsfreq) * self.cfg["autopw"][
                                           "peak_selection_cutoff_sig"],
                                       color="blue", label="Minimum Selection Amplitude", linestyle="--")
            self.add_pg_formatting_to_curfig(legend=True)
            pl.savefig(f"{self.output_dirs['pgs_slf']}/pg{n}.png")
            pl.clf()
        # lopoly =======================================
        if c_pg.log_polypar is not None:
            self.add_pg_plot_to_curfig(x=c_pg.lsfreq, y=c_pg.lsamp, label="Data")
            self.add_pg_plot_to_curfig(x=c_pg.lsfreq, y=c_pg.eval_poly_model(c_pg.lsfreq), color="red",
                                       label="LOPoly Fit")
            self.add_pg_formatting_to_curfig(legend=True)
            pl.savefig(f"{self.output_dirs['pgs_lopoly']}/pg{n}.png")
            pl.clf()



    def add_lc_plot_to_curfig(self, x, y, color="black", marker=".", alpha=1.0, linestyle="none", label="",
                              markersize=1):
        """
        Add a single light curve to the current plot. Passes through color, marker, alpha, linestyle, label,
        and markersize to pl.plot. Does not show or save the plot.
        """
        if self.cfg["output"]["plot_output_in_mmag"]:
            pl.plot(x, flux2mag(y, self.cfg["output"]["reference_flux"], self.cfg["output"]["reference_mag"]) * 1000,
                    color=color, marker=marker, alpha=alpha, linestyle=linestyle, label=label, markersize=markersize)
        else:
            pl.plot(x, y,
                    color=color, marker=marker, alpha=alpha, linestyle=linestyle, label=label, markersize=markersize)

    def add_lc_formatting_to_curfig(self, legend=False):
        """
        Adds formatting for a light curve to the current plot.
        :param bool legend: If true, add a legend
        :return: None
        """

        #%todo: set lc formatting to properly parse label configuration
        pl.xlabel(self.cfg["plots"]["lightcurves"]["x_label"])
        pl.ylabel(self.cfg["plots"]["lightcurves"]["y_label"])
        if legend:
            pl.legend()
        pl.tight_layout()

    def add_pg_plot_to_curfig(self, x, y, color="black", alpha=1.0, label="", linestyle="-"):
        """
        Adds a single periodogram to the current plot. Passes through color, alpha, label, and linestyle to pl.plot().
        Does not show or save the plot.
        """
        if self.cfg["output"]["plot_output_in_mmag"]:
            pl.plot(x, flux2mag(y, self.cfg["output"]["reference_flux"], self.cfg["output"]["reference_mag"]) * 1000,
                    color=color, alpha=alpha, label=label, linestyle=linestyle)
        else:
            pl.plot(x, y,
                    color=color, alpha=alpha, label=label, linestyle=linestyle)

    def add_pg_formatting_to_curfig(self, legend=False):
        """
        Adds formatting for a periodogram to the current plot.
        :param bool legend: If true, add a legend
        :return: None
        """
        # %todo: set pg formatting to properly parse label configuration
        pl.xlabel(self.cfg["plots"]["periodograms"]["x_label"])
        pl.ylabel(self.cfg["plots"]["periodograms"]["y_label"])
        if legend:
            pl.legend()
        pl.tight_layout()

    def get_freq_params_in_mmag(self, freq):
        """
        Get the amplitude in mmag for a frequency with its amplitude measured in flux, and the appropriately-adjusted
        phase. The conversion to magnitude flips the shape of the variability component described by the freq object,
        therefore this multiplies the amplitude in magnitude by -1 and adjusts the phase appropriately.
        :param freq: A frequency object
        :return: Amplitude in mmag, amplitude uncertainty in mmag, adjusted phase, adjusted initial magnitude in mmag
        """
        a, a_err, p, a0, p0 = freq.a, freq.sigma_a, freq.p, freq.a0, freq.p0
        am, a_errm = flux2mag_e(a, self.cfg["output"]["reference_flux"], self.cfg["output"]["reference_mag"], a_err)
        a0m = flux2mag(a0, self.cfg["output"]["reference_flux"], self.cfg["output"]["reference_mag"])
        # flip so it's positive
        am *= -1
        # adjust phase appropriately
        pm = p - 0.5
        p0m = p0 - 0.5
        if pm < 0:
            pm += 1
        if p0m < 0:
            p0m += 1
        return am * 1000, a_errm * 1000, pm, a0m * 1000

    def save_freqs(self, freqs: FrequencyContainer):
        """
        Format and save a frequency list as a csv
        :param freqs: A list of Freq objects to save
        :return: None
        """
        with open(f"{self.output_dirs['base']}/frequencies.csv", "w") as f:
            f.write(f"Freq,Amp,Phase,Freq_err,Amp_err,Phase_err,Sig_slf,Sig_lopoly,Sig_avg\n")
            if self.cfg["output"]["plot_output_in_mmag"]:
                for freq in freqs.get_flist():
                    am, sigma_am, pm, _ = self.get_freq_params_in_mmag(freq)
                    f.write(
                        f"{freq.f},{am},{pm},{freq.sigma_f},{sigma_am},{freq.sigma_p},{freq.sig_slf},{freq.sig_poly},{freq.sig_avg}\n")
            else:
                for freq in freqs.get_flist():
                    f.write(
                        f"{freq.f},{freq.a},{freq.p},{freq.sigma_f},{freq.sigma_a},{freq.sigma_p},{freq.sig_slf},"
                        f"{freq.sig_poly},{freq.sig_avg}\n")

    def save_freqs_as_latex_table(self, freqs: FrequencyContainer):
        """
        Format and save a frequency list as a latex table
        :param freqs: A list of Freq objects to save
        :param path: The path at which to save the data
        :return: None
        """
        with open(f"{self.output_dirs['auxiliary']}/frequencies.csv", "w") as f:
            for freq in freqs.get_flist():
                am, a_errm, pm, _ = self.get_freq_params_in_mmag(freq)
                f.write(f"{freq.n + 1} & {format_output(freq.f, freq.sigma_f, 2)} & {format_output(am, a_errm, 2)} &"
                        f" {format_output(pm, freq.sigma_p, 2)} & {round(freq.sig_slf, 2)} & {round(freq.sig_avg, 2)} &"
                        f"\\\\ \n")

    def save_lc(self, lightcurve, path):
        """
        Save a Lightcurve object at the specified path as a space-delimited 3-column file
        """
        np.savetxt(path, X=np.transpose([lightcurve.time, lightcurve.data, lightcurve.err]), delimiter=" ")

    def save_pg(self, periodogram, path):
        """
        Save a Periodogram object at the specified path as a space-delimited 2-column file
        """
        np.savetxt(path, X=np.transpose([periodogram.lsfreq, periodogram.lsamp]), delimiter=" ")



from astropy.timeseries import LombScargle
from typing import Union
import numpy as np
from scipy.optimize import curve_fit

import pywhiten
from pywhiten.optimization.models import n_model_poly, slf_noise
import matplotlib.pyplot as pl


class Periodogram:
    """
    Class which stores and operates on a Lomb-Scargle periodogram.

    Attributes:
        lsfreq (ndarray): frequency axis of periodogram
        lsamp (ndarray): amplitude axis of periodogram
        slf_p (list): Parameter list describing stochasic low frequency model fit - Formatted as
        [x0, alpha_0, gamma, Cw]. See Bowman et al. (2019) for details on these parameters.
        slf_p_err (list): List of diagonal elements of the covariance matrix for the slf noise fit. Estimates of the
            uncertainties on the parameters contained in slf_p.
        p_approx_nyquist_f (float): The nyquist frequency assuming even sampling across the time baseline. This will NOT
            be the nyquist frequency for data with gaps, but it's the best way to handle the general case
            for automatically generating periodogram boundaries
        p_resolution (float): the minimum separation necessary to reliably resolve two periodic components close in
            frequency using a Lomb-Scargle-based pre-whitening methodology.
        log_polypar (ndarray): a list of coefficients for a polynomial fit in log-log space.
    """
    lsfreq = None
    lsamp = None
    log_polypar = None
    slf_p = None
    slf_p_err = None
    p_approx_nyquist_f = None
    p_resolution = None
    cfg = None

    def __init__(self, time: np.ndarray, data: np.ndarray, lsfreq: Union[str, np.ndarray] = "auto",
                 fbounds = None,
                 pts_per_res: int = pywhiten.cfg.default_cfg["periodograms"]["points_per_resolution_element"],
                 cfg=None):
        """
        Constructor for Periodogram class
        Args:
            time (ndarray): Time axis of time series
            data (ndarray): Data axis of time series
            lsfreq (str, ndarray): Frequency grid to compute the periodogram over. Default behaviour is "auto" which
                automatically generates a grid. Alternatively, an array of floats may be provided directly.
            fbounds (tuple): bounds for periodogram in frequency of format (lower, upper)
            pts_per_res (int): number of points to generate per resolution element. Ignored if fgrid is provided.
            cfg (dict): A configuration dict. If nothing is provided, the default is used.
        Returns:
            Nothing
        Raises:
            TypeError: If input values aren't matched to the allowable types
        """
        if cfg is not None:
            self.cfg = cfg
        else:
            self.cfg = pywhiten.cfg.default_cfg
        # frequencies within this value are indistinguishable in this type of analysis
        self.p_resolution = 1.5 / (max(time) - min(time))
        # approximate nyquist frequency
        self.p_approx_nyquist_f = len(time) / (2 * (max(time) - min(time)))

        # set up a frequency grid. This is the most configurable component so it has a few branching paths for setup.
        if lsfreq == "auto":
            if fbounds is not None:  # explicitely provided bounds
                ll, ul = fbounds[0], fbounds[1]
            else:  # bounds not explicitely provided so they must be generated
                ll = self.cfg["periodograms"]["lower_limit"]
                if ll == "resolution":
                    ll = self.p_resolution
                ul = self.cfg["periodograms"]["upper_limit"]
            self.lsfreq = np.linspace(ll, ul, pts_per_res * int(abs(ll-ul)/(self.p_resolution)))
        else:
            self.lsfreq = lsfreq
        # generates pseudo-power spectrum
        lspower = LombScargle(time, data, normalization="psd").power(self.lsfreq)
        # normalizes the pseudo-power spectrum to an amplitude spectrum
        self.lsamp = 2 * (abs(lspower) / len(time)) ** 0.5

    def highest_ampl(self, excl_mask: np.ndarray = None):
        """
        Gets the frequency and amplitude of the highest peak in the periodogram
        Args:
            excl_mask (ndarray): A boolean array specifying indices (in lsfreq/lsamp) to ignore
        Returns:
            float: The frequency corresponding to the maximum amplitude on the periodogram
            float: The maximum amplitude of the periodogram
        """
        if excl_mask is None:
            excl_mask = np.ones(len(self.lsfreq), dtype=bool)
        filtered_lsamp = self.lsamp[excl_mask]
        filtered_lsfreq = self.lsfreq[excl_mask]

        # if the exclusion mask excludes all values from selection
        if len(filtered_lsamp) == 0:
            return None, None

        ymax = np.argmax(filtered_lsamp)
        xatpeak = filtered_lsfreq[ymax]
        return xatpeak, filtered_lsamp[ymax]

    def find_troughs(self, center):
        """
        Determine local minima on either side of a specified value
        Args:
            center (float): specified value

        Returns:
            int: index of lower local minimum (to the left of the peak)
            int: index of upper local minimum (to the right of the peak)
        """
        count = 2
        left_i, right_i = None, None
        while (not left_i) or (not right_i):
            count += 1
            # check for left trough
            if not left_i and (center - count == 0 or self.lsamp[center - count] >= self.lsamp[center - count + 1]):
                left_i = center - count + 1
            if not right_i and (
                    center + count == len(self.lsfreq) or self.lsamp[center + count] >= self.lsamp[center + count - 1]):
                right_i = center + count - 1
        if left_i < 0:  # ensure no wraparound
            left_i = 0
            if right_i > len(self.lsfreq):
                left_i = len(self.lsfreq)
        return left_i, right_i

    def find_index_of_freq(self, t):
        """
        Finds the index of lsfreq holding the closest frequency to a specified frequency
        Args:
            t (): a specified frequency

        Returns:
            int: the index of lsfreq holding the closest frequency to t
        """
        # just do linear search
        # this could be a binary search but it's not used all that often
        for i in range(len(self.lsfreq)-1):
            # check if t is between values at i, i+1
            if self.lsfreq[i] <= t < self.lsfreq[i+1]:
                # find out which one it is closer to
                if abs(self.lsfreq[i] - t) > abs(self.lsfreq[i+1] - t):
                    return i
                else:
                    return i+1
        print("[pywhiten][WARNING]: attempted to retrieve index of frequency outside of Periodogram lsfreq span")
        if t < self.lsfreq[0]:
            return 0
        else:
            return len(self.lsfreq)

    def peak_sig_box(self, center_val_freq, freq_amp, bin_r=1.0):
        """
        Gets the significance of a peak by considering the average periodogram value around the peak. Should NOT be used
        to measure significance in a residual periodogram where the peak has already been extracted (use sig_box()
        instead)
        Args:
            center_val_freq (float): frequency corresponding to maximum amplitude of peak
            freq_amp (float): amplitude of peak
            bin_r (float): radius in frequency to average. If this is less than the width of the peak

        Returns:
            float: A measured significance for the input peak
        """
        center_i_freq = np.where(self.lsfreq == center_val_freq)[0][0]
        trough_left_i, trough_right_i = self.find_troughs(center_i_freq)
        # determine target frequencies on either side of peak
        lower_val_freq = self.lsfreq[center_i_freq] - bin_r
        upper_val_freq = self.lsfreq[center_i_freq] + bin_r

        if lower_val_freq > self.lsfreq[trough_left_i] or upper_val_freq < self.lsfreq[trough_right_i]:
            raise AverageRadiusTooNarrow("Peak width on at least one side exceeds the specified average radius.")

        # don't permit target frequencies outside the range spanned by lsfreq
        if lower_val_freq < 0:
            lower_val_freq = 0
        if upper_val_freq > max(self.lsfreq):
            upper_val_freq = max(self.lsfreq)

        # convert target frequencies to frequency indices
        lower_i_freq = self.find_index_of_freq(lower_val_freq)
        upper_i_freq = self.find_index_of_freq(upper_val_freq)

        if lower_i_freq < 0:
            lower_i_freq = 0
        if upper_i_freq > len(self.lsfreq):
            upper_i_freq = len(self.lsfreq)

        # average the lower and upper region then return
        lower_avg_region = self.lsamp[lower_i_freq:trough_left_i]
        upper_avg_region = self.lsamp[trough_right_i:upper_i_freq]
        total_avg_region = np.concatenate((lower_avg_region, upper_avg_region))
        avg_regions_avg = np.mean(total_avg_region)
        return freq_amp / avg_regions_avg

    def sig_box(self, center_val_freq: Union[float, np.ndarray], freq_amp: Union[float, np.ndarray], bin_r=1.0):
        """
        Find the significance of a periodic component model by comparing its amplitude against the average residual
        periodogram near its frequency.
        Args:
            center_val_freq (float): The frequency of the periodic component
            freq_amp (float): The amplitude of the periodic component.
            bin_r (float): The radius in frequency to average around center_val_freq
        Returns:
            float: A measured significance for the input periodic component
        """
        center_i_freq = self.find_index_of_freq(center_val_freq)
        # find values of edge frequencies
        lower_val_freq = self.lsfreq[center_i_freq] - bin_r
        upper_val_freq = self.lsfreq[center_i_freq] + bin_r
        # find left and right indices
        lower_i_freq = self.find_index_of_freq(lower_val_freq)
        upper_i_freq = self.find_index_of_freq(upper_val_freq)
        total_avg_region = self.lsamp[lower_i_freq:upper_i_freq]
        return freq_amp / np.mean(total_avg_region)

    def fit_lopoly(self, poly_order: int = pywhiten.cfg.default_cfg["periodograms"]["polyfit_order"]):
        """
        Fits the periodogram with a polynomial in log-log space and stores the coefficients in the
        attribute log_polypar
        Args:
            poly_order (int): The polynomial order to use in the model

        Returns:
            Nothing
        """
        p0 = np.ones(poly_order + 1)
        log10_lsfreq = np.log10(self.lsfreq)
        log10_lsamp = np.log10(self.lsamp)
        p, pcov = curve_fit(f=n_model_poly, xdata=log10_lsfreq, ydata=log10_lsamp, p0=p0)
        self.log_polypar = p

    def sig_poly(self, center_val_freq: float, freq_amp: float):
        """
        Find the significance of a periodic component model by comparing its amplitude against a low-order polynomial
        fit to the log-log residual periodogram. Conducts an order-5 polynomial fit if a polynomial fit hasn't yet
        been conducted on this periodogram.
        Args:
            center_val_freq (float, ndarray): The frequency of the periodic component
            freq_amp (float, ndarray): The amplitude of the periodic component
        Returns:
            float: A measured significance for the input periodic component
        """
        if self.log_polypar is None:
            self.fit_lopoly(5)
        logfreq = np.log10(center_val_freq)
        logval = n_model_poly(logfreq, *self.log_polypar)
        return freq_amp / (10 ** logval)

    def fit_slf(self):
        """
        Fits the periodogram with a red noise + white noise model from Bowman et al. (2019). Stores the parameters in
        the attribute slf_p, and parameter uncertainties in slf_p_err
        Returns:
            Nothing
        """
        p0 = [0.5, np.mean(self.lsamp), 0.5, 0]
        p, covar = curve_fit(slf_noise, xdata=self.lsfreq, ydata=self.lsamp, p0=p0)
        self.slf_p = p
        self.slf_p_err = np.array([covar[i, i] for i in range(len(p))])

    def sig_slf(self, center_val_freq: Union[float, np.ndarray], freq_amp: Union[float, np.ndarray]):
        """
        Find the significance of a periodic component model by comparing its amplitude against a low-order polynomial
        fit to the log-log residual periodogram. Conducts an order-5 polynomial fit if a polynomial fit hasn't yet
        been conducted on this periodogram.
        Args:
            center_val_freq (float, ndarray): The frequency of the periodic component
            freq_amp (float, ndarray): The amplitude of the periodic component
        Returns:
            float: A measured significance for the input periodic component
                """
        if self.slf_p is None:
            self.fit_slf()
        model_at_val = self.eval_slf_model(center_val_freq)
        return freq_amp / model_at_val

    def select_peak(self, method: str = "highest", min_prov_sig: float = 4.0, mask: np.ndarray = None):
        """
        Determines a frequency-amplitude pair suitable for a pre-whitening iteration. Note: 'avg' method currently not
        implemented and just returns highest.
        Args:
            method (string): Must be one of ['highest', 'slf', 'poly']. 'highest' selects the frequency-amplitude
                pair strictly according to the highest amplitude in the periodogram. Each of the 'slf', 'poly',
                 methods impose an additional significance criterion that the highest peak must exceed, and which
                of the three chosen specifies how the significance is measured.
            min_prov_sig (float): The minimum significance that a prospective peak must exceed if using the "slf",
                "poly" methods.
            mask (ndarray): Array of boolean values excluding regions from peak selection. These regions are still used
                to measure significances, if applicable.
            cutoff (int): The number of times to try finding a peak above the provisional significance level before
                aborting and returning Nones. Only applies to box_avg selecction.
        Returns:
            float: the selected frequency. None if method is set to one of "slf", "poly" and a peak exceeding
                min_prov_sig cannot be found. Also returns None if user provides a mask consisting of all False values
            float: the selected amplitude. None if method is set to one of "slf", "poly" and a peak exceeding
                min_prov_sig cannot be found. Also returns None if user provides a mask consisting of all False values
        """
        # set up a mask to exclude
        working_mask = np.ones(len(self.lsfreq), dtype=bool)

        if mask is not None:
            working_mask*=mask

        if method == "highest":
            pass
        elif method == "slf":
            # we can just use the fit functions to dettermine a provisional noise level for all lsfreq values,
            # then use this to make a mask to pass to highest_ampl
            working_mask *= self.sig_slf(self.lsfreq, self.lsamp) > min_prov_sig
            if self.cfg["output"]["debug"]["show_peak_selection_plots"]:
                self.peak_selection_debug_plot_slf(working_mask)
        elif method == "poly":
            # we can just use the fit functions to dettermine a provisional noise level for all lsfreq values,
            # then use this to make a mask to pass to highest_ampl
            working_mask *= self.sig_poly(self.lsfreq, self.lsamp) > min_prov_sig
        else:
            raise ValueError(f"method={method} in select_peak not in the allowable options: highest, slf,"
                                     f"poly, avg")

        if (~working_mask).all():  # this condition checks if the mask contains all false values
            return None, None
        else:
            return self.highest_ampl(excl_mask=working_mask)
    def eval_slf_model(self, x):
        """Performs a SLF fit if one hasn't been performed already,
        and evaluates the SLF model fit at the x value(s) provided."""
        if self.slf_p is None:
            self.fit_slf()
        return slf_noise(x, *self.slf_p)

    def eval_poly_model(self, x):
        if self.log_polypar is None:
            self.fit_lopoly(self.cfg["periodograms"]["polyfit_order"])

        logfreq = np.log10(self.lsfreq)
        logvals = n_model_poly(logfreq, *self.log_polypar)
        return 10**logvals


    def debug_plot(self):
        pl.plot(self.lsfreq, self.lsamp, color="black")
        pl.show()
        pl.clf()
    def peak_selection_debug_plot_slf(self, working_mask):
        pl.plot(self.lsfreq[working_mask], self.lsamp[working_mask], color="green", linestyle="none", marker=".", markersize=2)
        pl.plot(self.lsfreq[~working_mask], self.lsamp[~working_mask], color="black", linestyle = "none", marker=".", markersize=2)
        pl.plot(self.lsfreq, slf_noise(self.lsfreq, *self.slf_p)*3, color="red", linestyle = "--")
        pl.plot(self.lsfreq, slf_noise(self.lsfreq, *self.slf_p), color="red")
        pl.axhline(0)
        pl.xlim(0, 10)
        pl.show()
        pl.clf()





class AverageRadiusTooNarrow(Exception):
    pass

if __name__=="__main__":
    time = np.linspace(0, 100, 100)
    data = np.random.rand(100)


"""
A module for handling pre-whitening input and output

This includes some utility functions for converting between flux and magnitude as well as a function for formatting
output measurements with uncertainties in the common 1.23(1) format. The OutputManager class provides functionality for
setting up an output structure, making consistent lightcurve and periodogram plots, and saving output and auxiliary
data
Classes:
    OutputManager

Functions:
    flux2mag
    flux2mag_e
    format_output
"""

from pywhiten.pwio.OutputManager import OutputManager
from pywhiten.pwio.utilities import *


import tomli
import os
from pywhiten.pwio.OutputManager import OutputManager
import shutil
def test_OutputManager_tree_setup():
    default_config = os.getcwd()+"/pywhiten/cfg/default.toml"
    with open(default_config, "rb") as f:
        cfg = tomli.load(f)

    test_output_manager = OutputManager(cfg=cfg)
    if os.path.exists(os.getcwd()+"/pw_out"):
        shutil.rmtree(os.getcwd()+"/pw_out")
        return 0
    else:
        return 1


from typing import Union
from pywhiten.optimization.models import sin_model

class Frequency:
    """
    An object storing a set of parameters for a sinusoidal single-frequency model.
    Attributes:
        f : float
            Model frequency
        sigma_f : float
            frequency uncertainty
        a : float
            Model amplitude
        sigma_a : float
            amplitude uncertainty
        p : float
            Model phase
        sigma_p : float
            phase uncertainty
        t0 : float
            The timestamp which is used as the reference for the measurement of phase. The phase measurement is
            meaningless without this specified.
        sig_poly : float
            Significance as measured using a low-order polynomial fit to the residual light curve's periodogram
        sig_avg : float
            Significance as measured using a rolling box average of the residual light curve's periodogram
        sig_slf : float
            Significance as measured by fitting the Bowman et al. SLF variability model to the residual light curve's
            periodogram

        f0 : float
            The value of f when this object was initialized
        a0 : float
            The value of a when this object was initialized
        p0 : float
            The value of p when this object was initialized

        n: int
            An index label
        model_function : function
            A function of the form f(x, f, a, p) which described the variability model.
    """
    f : float
    sigma_f : float
    a : float
    sigma_a : float
    p : float
    sigma_p : float
    t0 : float
    sig_poly : float
    sig_avg : float
    sig_slf : float
    f0 : float
    a0 : float
    p0 : float

    n = None
    model_function = None

    def __init__(self, f:float, a:float, p:float, t0:float, model_function=sin_model, n:Union[int, None]=None):
        """

        Args:
            f (float): initial frequency
            a (float): initial amplitude
            p (float): initial phase
            t0 (float): reference time for phase
            model_function (function reference): A model function. This will typically be the model used to opimize the
                f, a, p parameters. This is required to fully specify the meaning of f, a, p, and t0. Must be of the
                form f(x, f, a, p). Defaults to sin model, as with anything else that utilizes a model of this form.
            n (int): index label
        Returns:
            Nothing
        """
        self.f = f
        self.f0 = f
        self.a = a
        self.a0 = a
        self.p = p
        self.p0 = p
        self.t0 = t0
        self.n = n
        self.model_function = model_function

        self.sig_poly = 0
        self.sig_avg = 0
        self.sig_slf = 0


    def update(self, f, a, p):
        """Updates the stored frequency, amplitude, and phase"""
        self.f, self.a, self.p = f, a, p

    def get_parameters(self):
        """Returns the frequency, amplitude, phase"""
        return self.f, self.a, self.p

    def adjust_params(self):
        """Adjusts amplitude to be positive and phase to be within [0, 1] without changing the model."""
        # adjust amplitude
        if self.a < 0:
            self.a = abs(self.a)
            self.p += 0.5

        if self.p > 1 or self.p < 0:
            self.p = self.p % 1

    def evaluate_model(self, x):
        return self.model_function(x, self.f, self.a, self.p)

    def prettyprint(self):
        """Prints the parameters to console"""
        if self.n is not None:
            print(f"\tf{self.n}: f = {self.f:.5f} | a = {self.a:.3f} | phi = {self.p:.3f}")
        else:
            print(f"\tf(no idx.): f = {self.f:.5f} | a = {self.a:.3f} | phi = {self.p:.3f}")



"""
pywhiten is a python package for conducting Lomb-Scargle-based pre-whitening to identify
frequencies, amplitudes, and phases of sinusoidal variability signals in time series data.

The basic methodology implemented in this frequency analysis package is based on least-squares fitting of sinusoidal
models to time series data (based on the seminal papers of N. R. Lomb (1976) and J. D. Scargle (1982). This is a
popular approach to frequency analysis within the observational astronomy community, however other approaches have seen
a significant degree of use (e.g. Phase Dispersion Minimization) as well. The approach implemented in this package has
the advantages of 1) being computationally fast, utilizing the Lomb-Scargle Periodogram and Levenberg-Marquardt
minimization routines, and 2) providing results in an easily-quantifiable form. The core functionality of this program
was written and validated by Erik William Stacey as a principle component of the analysis presented in his MSc
thesis at the Royal Military College of Canada.

Classes:
    Pywhitener
Modules:
    data: Contains data structures that are useful for pre-whitening
    optimization: Contains all mathematical optimization functionality for the package
    pwio: Contains input/output functionality
Variables:
    default_cfg (dict): A dictionary storing the default configuration
    pkg_path (string): A string storing the local path to the package
"""
import pywhiten.cfg
import pywhiten.data
import pywhiten.pwio
import pywhiten.optimization
from pywhiten.PyWhitener import PyWhitener












import tomli

import pywhiten
from pywhiten.data import *
from pywhiten.pwio import OutputManager
from pywhiten.optimization.Optimizer import Optimizer
import os
import numpy as np


class PyWhitener:
    """
    The main class for conduction pre-whitening analyses using pywhiten.
    Attributes:
        lcs (list): A list of Lightcurve objects. The first entry is populated with the constructor, and subsequent
            entries are added when pre-whitening
        freqs (FrequencyContainer): A container of Frequency objects identified while pre-whitening
        cfg (dict): A nested dictionary representing the configuration supplied in [package path]/cfg/default.toml,
            or however manually specified
        optimizer (Optimizer): A class for handling the optimization steps of the pre-whitening process
    """
    lcs: list
    freqs: FrequencyContainer
    cfg: dict
    output_manager: OutputManager
    optimizer: Optimizer

    def __init__(self, time, data, err=None, cfg_file="default", cfg=None):
        """

        Args:
            time (np.ndarray): Time series time axis
            data (np.ndarray): Time series data axis
            err (np.ndarray): Time series data point weights, in the form of uncertainties
            cfg_file (string): Path to (or filename, if file is in current working directory or cfg directory) a config
                file, which will be used to overwrite the default configuration. Does not need to be fully populated,
                only the populated fields will overwrite.
            cfg (dict): A nested dict which mimics the structure of the configuration file. This can also be used to
                manually override the configuration entries. For example, to set the periodogram x labels the following
                should return the desired value: example_dict['output']['paths']['pg_x_label']
        """
        # using this method to load the cfg maintains backwards compatibility with python 3.7 in the least painful form
        # load default cfg, then overwrite with config file specified by cfg
        pkg_path = pywhiten.pkg_path
        default_config = pkg_path + "/cfg/default.toml"
        with open(default_config, "rb") as f:
            self.cfg = tomli.load(f)

        if cfg_file != "default" and type(cfg_file) == str:
            # if specified cfg file is in current directory, use that
            if os.path.exists(f"{os.getcwd()}/+{cfg_file}"):
                with open(f"{os.getcwd()}/+{cfg_file}", "rb") as f:
                    loaded_cfg = tomli.load(f)
                    self.cfg = merge_dict(self.cfg, loaded_cfg)
                self.cfg["title"] = "Config loaded from file in working directory"
            # try to find it in the cfg directory instead
            elif os.path.exists(f"{pkg_path}/cfg/{cfg_file}"):
                with open(f"{pkg_path}/cfg/{cfg_file}", "rb") as f:
                    loaded_cfg = tomli.load(f)
                    self.cfg = merge_dict(self.cfg, loaded_cfg)
                self.cfg["title"] = "Config loaded from file in package cfg directory"
            # finally, check to see if the cfg file was specified as a full path
            elif os.path.exists(cfg_file):
                with open(cfg_file, "rb") as f:
                    loaded_cfg = tomli.load(f)
                    self.cfg = merge_dict(self.cfg, loaded_cfg)
                self.cfg["title"] = "Config loaded from file at specified path"
        # use the cfg argument of the initializer to overwrite any cfg entries
        if cfg is not None and type(cfg) == dict:
            self.cfg = merge_dict(self.cfg, cfg)
            self.cfg["title"] += " - Merged with runtime-specified arguments"
        # handle special logic cases
        pass
        # config setup done
        # now set up initial light curve, frequency container, output mgr
        if self.cfg["input"]["subtract_mean"]:
            self.lcs = [Lightcurve(time, data-np.mean(data), err, cfg=self.cfg)]
        else:
            self.lcs = [Lightcurve(time, data, err, cfg=self.cfg)]
        self.freqs = FrequencyContainer()
        self.output_manager = OutputManager(cfg=self.cfg)
        self.optimizer = Optimizer(cfg = self.cfg)

        # we're now ready to do some frequency analysis

    def id_peak(self, method, min_prov_sig = 0, idx=-1):
        """
        Gets a candidate frequency/amplitude pair from a periodogram belonging to a light curve in lcs list
        Args:
            method (str): A method used to identify a peak. Allowed values: ['highest', 'slf', 'poly']. Refer
                to data.Periodogram.select_peak() for more details
            idx (int): index in lcs list to identify a peak from

        Returns:
            All values will be None if entire periodogram is excluded from selection. This can occur if using the 'slf'
            'poly' or 'avg' methods and no points in the periodogram meet the required minimum significance. In this
            case, this is utilized as a signal to trigger a stop criterion when automatically pre-whitening.
            float: candidate frequency
            float: candidate amplitude.
        """
        return self.lcs[idx].periodogram.select_peak(method=method,
                                                     min_prov_sig=min_prov_sig)

    def it_pw(self, peak_selection_method="highest"):
        """
        Conducts a single pre-whitening iteration. Identifies a candidate frequency/amplitude from the most recent
        light curve, fits a single-frequency sinusoid to the most recent light curve and includes it as a frequency,
        improves all parameters of all frequencies in multi-frequency fit against original light curve, and makes a new
        light curve ready for another iteration.
        Args:
            peak_selection_method (str):  the peak selection method used to get a candidate frequency/amplitude. can be
                one of ['highest', 'slf', 'poly'] - see id_peak method of data.Periodogram class

        Returns:
            int : a flag indicating whether the PW iteration succeeded
                0 if iteration was successful
                1 if peak identification failed using the specified method (which is the termination criterion
                    for autopw)
                2 if a new lightcurve couldn't be generated. This probably happens if
                    autopw.new_lc_generation_method is set incorrectly in cfg
        """
        if self.cfg["output"]["print_runtime_messages"]:
            print(f"[pywhiten] ITERATION {self.freqs.n + 1}")

        # First stage, identify a candidate frequency/amplitude from the most recent periodogram
        candidate_frequency, candidate_amplitude = self.id_peak(method=peak_selection_method,
                                                                min_prov_sig=
                                                                self.cfg["autopw"]["peak_selection_cutoff_sig"])
        if self.cfg["output"]["debug"]["print_debug_messages"]:
            print(f"[DEBUG][pywhiten] Identified candidate frequency {candidate_frequency} and"
                  f" amplitude {candidate_amplitude} using method {peak_selection_method}")
        if candidate_frequency is None and candidate_amplitude is None:
            if self.cfg["output"]["debug"]["print_debug_messages"]:
                print(f"[DEBUG][pywhiten] TERMINATION CRITERION SATISFIED")
            return 1

        # Second stage, conduct a single-frequency fit
        sf_f, sf_a, sf_p, sf_model = self.optimizer.single_frequency_optimization(self.lcs[-1].time,
                                                                         self.lcs[-1].data,
                                                                         self.lcs[-1].err,
                                                                         candidate_frequency, candidate_amplitude, 0.5)
        self.freqs.add_frequency(Frequency(f=sf_f, a=sf_a, p=sf_p, model_function=self.optimizer.sf_func,
                                           t0=self.cfg["input"]["input_t0"], n=len(self.freqs.get_flist())))
        if self.cfg["output"]["print_runtime_messages"]:
            print(f"[pywhiten] Identified single frequency model:")
            self.freqs.get_flist()[-1].prettyprint()

        # Third stage, conduct the multi-frequency fit
        mf_mod = self.optimizer.multi_frequency_optimization(self.lcs[0].time,
                                                             self.lcs[0].data,
                                                             self.lcs[0].err,
                                                             self.freqs.get_flist())

        self.output_manager.save_it(self.lcs, self.freqs, self.optimizer.c_zp)
        if self.cfg["output"]["print_runtime_messages"]:
            print(f"[pywhiten] Completed optimization of current complete variability model:")
            for f_to_print in self.freqs.get_flist():
                f_to_print.prettyprint()

        # Final Stage, make a new light curve and append it to the lightcurves list
        if self.cfg["autopw"]["new_lc_generation_method"] == "mf":
            self.lcs.append(Lightcurve(self.lcs[0].time, self.lcs[0].data - mf_mod, err = self.lcs[0].err, cfg=self.cfg))
            return 0
        elif self.cfg["autopw"]["new_lc_generation_method"] == "sf":
            self.lcs.append(Lightcurve(self.lcs[-1].time, self.lcs[-1].data - sf_model, err=self.lcs[0].err, cfg=self.cfg))
            return 0
        return 2

    def auto(self):
        for i in range(self.cfg["autopw"]["cutoff_iteration"]):
            success_flag = -1
            if i < self.cfg["autopw"]["peak_selection_highest_override"]:
                success_flag = self.it_pw(peak_selection_method="highest")
            else:
                success_flag = self.it_pw(peak_selection_method=self.cfg["autopw"]["peak_selection_method"])
            if success_flag == 1:
                break
        self.post_pw()



    def post_pw(self, residual_lc_idx : int = -1):
        """
        Conducts post-pre-whitening tasks. Principally, computes significances and parameters uncertainties for
        frequencies
        Args:
            residual_lc_idx (int): The integer corresponding to the index of the residual light curve in self.lcs.
                Defaults to -1 (the most recent LC added)
        Returns:
            Nothing
        """
        self.freqs.compute_significances(self.lcs[residual_lc_idx].periodogram)
        self.freqs.compute_parameter_uncertainties(self.lcs[residual_lc_idx])
        self.output_manager.save_freqs(freqs=self.freqs)
        self.output_manager.save_freqs_as_latex_table(freqs=self.freqs)
        self.output_manager.save_it(self.lcs, self.freqs, self.optimizer.c_zp)

def merge_dict(old_dict, new_dict):
    """
    Merges two dictionaries, overwriting entries in old_dict with entries in new_dict if both contain entries with the
    same access pattern (key, or nested set of keys). Recursively merges consistutent dictionaries as well.
    Args:
        old_dict (dict): A possibly nested dictionary. Entries in this dictionary are overwritten in case of conflicts.
        new_dict (dict): A possibly nested dictionary

    Returns:
        dict: the merged dictionary
    """
    out_dict = {}
    for key in new_dict.keys():
        if key in old_dict.keys() and type(old_dict[key]) == dict:
            out_dict[key] = merge_dict(old_dict[key], new_dict[key])
        else:
            out_dict[key] = new_dict[key]
    for key in old_dict.keys():
        if key not in out_dict.keys():
            out_dict[key] = old_dict[key]
    return out_dict


if __name__ == "__main__":
    test_pywhitener = PyWhitener(np.linspace(0, 10, 100), np.random.rand(100), np.ones(100))
    print(test_pywhitener.cfg)



import pywhiten
import numpy as np
from pywhiten.data import Lightcurve

def test_FrequencyContainer_basic_functionality():
    f1 = pywhiten.data.Frequency(1.0, 2.0, 3.0, t0=0, n=0)
    f2 = pywhiten.data.Frequency(2.2, 2.3, 4.1, t0=0, n=1)
    f3 = pywhiten.data.Frequency(0.1, 20, 0.3, t0=0, n=2)

    test_container = pywhiten.data.FrequencyContainer(f1, f2)
    assert len(test_container.get_flist()) == 2
    assert test_container.get_last_frequency().f == 2.2
    test_container.add_frequency(f3)
    assert len(test_container.get_flist()) == 3
    assert test_container.get_last_frequency().f == 0.1

    test_model = test_container.mf_model(np.linspace(0, 10, 100), 0)
    assert np.all(test_model == pywhiten.optimization.n_sin_model(np.linspace(0, 10, 100),
                                                            1.0, 2.2, 0.1,
                                                            2.0, 2.3, 20,
                                                            3.0, 4.1, 0.3,
                                                            0))

def test_FrequencyContainer_significances_and_uncertainties():
    f1 = pywhiten.data.Frequency(1.0, 2.0, 3.0, t0=0, n=0)
    f2 = pywhiten.data.Frequency(2.2, 2.3, 4.1, t0=0, n=1)
    f3 = pywhiten.data.Frequency(0.1, 20, 0.3, t0=0, n=2)

    test_container = pywhiten.data.FrequencyContainer(f1, f2, f3)

    test_lc_time = np.linspace(0, 10, 1000)
    test_lc_data = (np.random.rand(1000)-0.5) * 20
    for i in range(1, 1000):
        test_lc_data[i] += test_lc_data[i-1]*0.95


    test_lc = Lightcurve(test_lc_time, test_lc_data-np.mean(test_lc_data), np.ones(1000))

    test_container.compute_parameter_uncertainties(test_lc)
    test_container.compute_significances(test_lc.periodogram)

    for i in range(len(test_container.get_flist())):
        assert test_container.get_flist()[i].sig_slf is not None
        assert test_container.get_flist()[i].sig_box is not None
        assert test_container.get_flist()[i].sig_poly is not None
        assert test_container.get_flist()[i].sigma_f is not None
        assert test_container.get_flist()[i].sigma_a is not None
        assert test_container.get_flist()[i].sigma_p is not None




import pywhiten

def test_Frequency():
    test_object_noidx = pywhiten.data.Frequency(1.2, 1.3, 0.3, 0)
    assert test_object_noidx.get_parameters() == (1.2, 1.3, 0.3)
    test_object_noidx.update(2.2, -2.3, 0.3)
    assert test_object_noidx.get_parameters() == (2.2, -2.3, 0.3)
    test_object_noidx.adjust_params()
    # accounts for floating point error. In reality we don't want to throw this out like this.
    test_object_noidx.p = round(test_object_noidx.p, 5)
    assert test_object_noidx.get_parameters() == (2.2, 2.3, 0.8)
    test_object_noidx.update(2.2, -2.3, 55.3)
    test_object_noidx.adjust_params()
    test_object_noidx.p = round(test_object_noidx.p, 5)
    assert test_object_noidx.get_parameters() == (2.2, 2.3, 0.8)

    # check to make sure we still have original parameters
    assert test_object_noidx.f0 == 1.2
    assert test_object_noidx.a0 == 1.3
    assert test_object_noidx.p0 == 0.3


    test_object_widx = pywhiten.data.Frequency(1.2, 1.3, 0.3, 0, n=1)


import os
import tomli

pkg_path = os.path.dirname(os.path.abspath(__file__))
default_cfg_path = os.path.join(pkg_path, 'default.toml')

with open(default_cfg_path, "rb") as f:
    default_cfg = tomli.load(f)

def make_config_file(path="./default.toml"):
    import shutil
    """
    Copies the default config file to somewhere (typically) outside the package files, so it can be edited by the user.
    Args:
        path (str): Path to copy the default configuration to

    Returns:

    """
    shutil.copyfile(default_cfg_path, path)
    print(f"[pywhiten] Made a copy of default configuration file at {path}!")


import os
import tomli
import numpy as np
from pywhiten.optimization.Optimizer import Optimizer
from pywhiten.optimization.models import sin_model, n_sin_model
from pywhiten.data.Frequency import Frequency
from pywhiten.data.FrequencyContainer import FrequencyContainer

def test_sin_model():
    test_x = np.linspace(0, 10, 1000)
    test_y = sin_model(test_x, 1.0, 10, 0.5)
    assert round(test_y[0], 5) == 0.00000

def test_Optimizer_sf():
    default_config = os.getcwd()+"/pywhiten/cfg/default.toml"
    with open(default_config, "rb") as f:
        cfg = tomli.load(f)

    test_Optimizer = Optimizer(cfg=cfg)

    assert test_Optimizer.sf_func == sin_model
    assert test_Optimizer.mf_func == n_sin_model
    assert test_Optimizer.c_zp == 0

    test_x = np.linspace(0, 10, 1000)
    test_y = sin_model(test_x, 1.0, 10, 0.5)

    fit_f, fit_a, fit_p, model = test_Optimizer.single_frequency_optimization(test_x, test_y, np.ones(len(test_x)), 1.5, 8, 0)

def test_Optimizer_mf():
    default_config = os.getcwd()+"/pywhiten/cfg/default.toml"
    with open(default_config, "rb") as f:
        cfg = tomli.load(f)

    test_Optimizer = Optimizer(cfg=cfg)

    assert test_Optimizer.sf_func == sin_model
    assert test_Optimizer.mf_func == n_sin_model
    assert test_Optimizer.c_zp == 0

    test_x = np.linspace(0, 10, 1000)
    test_y = sin_model(test_x, 1.0, 10, 0.5) + sin_model(test_x, 1.4, 7, 0.4)

    f1 = Frequency(1.1, 9.0, 0.7, t0=0, n=0)
    f2 = Frequency(1.6, 8.1, 0.8, t0=0, n=1)

    test_container = FrequencyContainer(f1, f2)
    assert test_container.get_flist()[0].f == 1.1
    model = test_Optimizer.multi_frequency_optimization(test_x, test_y, np.ones(len(test_x)),
                                                        test_container.get_flist())
    assert test_container.get_flist()[0].f != 1.1


from setuptools import setup, find_packages

setup(name="pywhiten",
      version="1.1.0",
      description="A python package for conducting Lomb-Scargle-based prewhitening of stellar time series. ",
      url="https://www.github.com/erikstacey/pywhiten",
      author="Erik William Stacey",
      author_email = "erik@erikstacey.com",
      license="MIT",
      packages= find_packages(where='.'),
      install_requires=[
            "numpy",
            "matplotlib",
            "scipy",
            "astropy",
            "lmfit",
            "tomli"
      ],
      tests_require=["pytest"],
      zip_safe = False,
      include_package_data=True,
      package_data={'': ['cfg/default.toml']},
      )


from pywhiten.data import Periodogram, Lightcurve
from pywhiten.data.Frequency import Frequency
from pywhiten.data.InvalidConstructorArgumentsError import InvalidContructorArgumentsError
from pywhiten.optimization.models import sin_model
import numpy as np

class FrequencyContainer:
    """
    An object for storing and operating on groups of frequencies

    Attributes:
        _flist : list
            List of Frequency objects
    """
    flist: list = []
    n = 0

    def __init__(self, *freqs):
        for i in range(len(freqs)):
            if type(freqs[i]) != Frequency:
                raise InvalidContructorArgumentsError("FrequencyContainer initialized with at least one argument not"
                                                      "of type Frequency")
        self.flist = [*freqs]
        self.n = len(self.flist)

    def get_flist(self):
        """
        Getter for frequency list, which shouldn't be direct accessed by user
        Returns:
            list: List of frequencies
        """
        return self.flist

    def mf_model(self, time, zp:float = 0):
        y = np.zeros(len(time))
        for f in self.flist:
            y += sin_model(time, *f.get_parameters())
        return y+zp

    def get_last_frequency(self):
        """
        Gets the frequency on the end of the flist
        Returns:
            Frequency: the most recently-added frequency in flist
        """
        return self.flist[-1]

    def add_frequency(self, freq:Frequency):
        self.flist.append(freq)
        self.n += 1

    def del_frequency(self, index):
        self.flist.pop(index)
        self.n -= 1

    def compute_significances(self, residual_periodogram:Periodogram, eval : tuple =("slf", "box", "poly")):
        """
        Computes significances for stored frequencies and stores them in class attributes for each frequency. Computes
        according to 3 methods and stores the results in separate class attributes by default.
        Args:
            residual_periodogram (Periodogram): A periodogram object, which is assumed to consist of noise in order to
                evaluate the signal to noise ratio of the frequencies. Typically, when conducting pre-whitening, the
                analyzer will remove all frequencies reasonably thought to be real and identifiable in the data then
                uses the residual periodogram to evaluate their confidence in their results.
            eval (tuple): A tuple of strings indicating which significance methods to use in the evaluation. By default,
                this calculates significances using a box average ("box), red noise model ("slf"), and low-order
                polynomial model ("poly") fit.

        Returns:
            Nothing
        """
        for i in range(len(self.flist)):
            if "slf" in eval:
                self.flist[i].sig_slf = residual_periodogram.sig_slf(self.flist[i].f, self.flist[i].a)
            if "box" in eval:
                self.flist[i].sig_box = residual_periodogram.sig_box(self.flist[i].f, self.flist[i].a)
            if "poly" in eval:
                self.flist[i].sig_poly = residual_periodogram.sig_poly(self.flist[i].f, self.flist[i].a)

    def compute_parameter_uncertainties(self, residual_light_curve:Lightcurve):
        """
        Compute uncertainties for frequency, amplitude, and phase parameters of stored frequencies. Uses Montgomery &
        Odonogue (1999) method, with Schwarzenberg-Czerny (multiple publications, see astronomer's guide to period
        searching, 2003) correction. This method is utilized in Degroote et al. (2009)
        Args:
            residual_light_curve (Lightcurve): A Lightcurve object, which is assumed to have no reasonably measurable
                frequency content remaining. Note that in many cases this may appear to have coherent variability,
                however this can arise from stochastic noise commonly seen in stellar light curves (see SLF noise).

        Returns:
            Nothing
        """
        N_eff = residual_light_curve.measure_N_eff()
        sigma_residuals = residual_light_curve.std()
        time_baseline = residual_light_curve.t_span()
        for i in range(len(self.flist)):
            self.flist[i].sigma_f = (6/N_eff) ** 0.5 * sigma_residuals / (np.pi * time_baseline * self.flist[i].a)
            self.flist[i].sigma_a = (2/N_eff) ** 0.5 * sigma_residuals
            self.flist[i].sigma_p = (2/N_eff) ** 0.5 * sigma_residuals / self.flist[i].a


import numpy as np

def flux2mag(data, ref_flux, ref_mag):
    """ Converts a data array with reference flux ref to magnitude"""
    return -2.51188643151 * np.log10(data/ref_flux) + ref_mag

def flux2mag_e(data, ref_flux, ref_mag, err):
    """ Converts a data array with reference flux ref to magnitude, including error propagation"""
    mag = -2.51188643151 * np.log10(data/ref_flux) + ref_mag
    err = (-2.51188643151/np.log(10))*err/data
    return mag, err

def format_output(n, e, npts):
    """
    Formats a measurement of the form 1.23 +/- 0.01 as 1.23(1).
    Args:
        n (float): a nominal value of a measurement
        e (float): an error corresponding to the nominal value
        npts (int): Number of digits in uncertainty to retain

    Returns:
        formatted string
    """
    split_n = str(n).split('.')
    split_e = str(e).split('.')
    roundto = 0
    if len(split_n) == 1: # int
        pass
    elif len(split_n) == 2: # float
        error_decimal = split_e[1]
        for i in range(len(error_decimal)):
            if error_decimal[i] != "0":
                roundto = i + npts
                while len(error_decimal) < roundto:
                    error_decimal += "0"
                break

        err_out = error_decimal[roundto-npts:roundto]
        return f"{n:.{roundto}f}({err_out})"

